from io import BytesIO
from typing import Any, List, Optional, Tuple

import httpx
from loguru import logger
from PIL import Image

from . import settings
from .data import Job, Online, Post, User

HEADERS = {
    "Connection": "keep-alive",
    "Accept-Language": "zh-CN,zh;q=0.9",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
    "Upgrade-Insecure-Requests": "1",
    "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3202.62 Safari/537.36",
}


class ApiException(Exception):
    """
    接口访问错误
    """
    def __init__(self, url: str, code: int, message: str, **kwargs):
        super().__init__(message)
        self.url = url
        self.code = code
        self.message = message
        self.kwargs = kwargs

    def __str__(self):
        return f"ApiException({self.url}, {self.code}, {self.message})"


class Session:
    """
    会话类
    
    基于 `httpx.AsyncClient`

    参考 `golang` 中错误处理方式
    """
    def __init__(self, url: str, headers: Optional[dict] = None) -> None:
        """
        Args:
            url: 基础接口地址
        
            headers: 自定义请求头
        """
        self.__session = httpx.AsyncClient(base_url=url)
        if headers is None:
            self.headers = HEADERS.copy()
        else:
            self.headers = headers

    def check_code(self, url: str, result: Any) -> Optional[ApiException]:
        if result["code"] != 0:
            return ApiException(url=url, **result)
        return None

    def data(self, result: dict) -> Any:
        return result.get("data")

    @property
    def session(self):
        return self.__session

    async def request(
        self,
        method: str,
        url: str,
        headers: Optional[dict] = None,
        *args,
        **kwargs
    ) -> Tuple[Any, Optional[Exception | ApiException]]:
        # prepare
        if headers is None:
            headers = self.headers
        # fetch
        try:
            if settings.DEBUG:
                logger.debug(f"\n{method} {url}\n  args: {args}\n  kwargs: {kwargs}\n  headers: {headers}")
            resp = await self.__session.request(method, url, headers=headers, *args, **kwargs)
            # if settings.DEBUG:
            #     logger.debug(resp.content)
            if resp.status_code != 200:
                return None, ApiException(url, resp.status_code, f"<Response [{resp.status_code}]>")
            result = resp.json()
        except Exception as e:
            return None, e
        # data
        err = self.check_code(url, result)
        if err is not None:
            return None, err
        return self.data(result), None

    async def get(self, url: str, headers: Optional[dict] = None, *args, **kwargs):
        return await self.request("GET", url, headers, *args, **kwargs)
    
    async def post(self, url: str, headers: Optional[dict] = None, *args, **kwargs):
        return await self.request("POST", url, headers, *args, **kwargs)
    
    async def must_get(self, url: str, headers: Optional[dict] = None, *args, **kwargs):
        return (await self.get(url, headers, *args, **kwargs))[0]
    
    async def must_post(self, url: str, headers: Optional[dict] = None, *args, **kwargs):
        return (await self.post(url, headers, *args, **kwargs))[0]
    

class Api(Session):
    """
    Api 实现层
    """
    def __init__(self, url: str, auth: str):
        """
        Args:
            url: 基础接口地址
        
            auth: 鉴权码
        """
        self.auth = auth
        Session.__init__(self, url, {"Authorization": auth})

    def set_auth(self, auth: str):
        """
        更新鉴权码
        """
        self.auth = auth
        self.headers["Authorization"] = auth

    async def version(self) -> str:
        """
        获取服务端版本
        """
        return await self.must_get("/version")

    async def list(self) -> Tuple[List[str], Optional[Exception | ApiException]]:
        """
        查看资源目录
        """
        return await self.get("/list")

    async def fetch(self, url: str) -> Tuple[bytes, Optional[ApiException]]:
        """
        解析图片网址并返回文件
        """
        resp = await self.session.get(f"/fetch/{url}")
        err = None if resp.status_code == 200 else ApiException(url, resp.status_code, f"<Response [{resp.status_code}]>")
        return resp.content, err

    @staticmethod
    def open(data: bytes) -> Image.Image:
        """
        打开图片
        """
        return Image.open(BytesIO(data))

    async def online(self) -> Tuple[Optional[Online], Optional[Exception | ApiException]]:
        """
        获取当前在线状态
        """
        resp, err = await self.get("/online")
        if err is not None:
            return None, err
        return Online(**resp), None

    async def token(self, uid: str) -> Tuple[dict, Optional[Exception | ApiException]]:
        """
        获取注册所需 Token
        """
        return await self.get("/token", params={"uid": uid})

    async def register(self, auth: str):
        """
        注册或获取鉴权码
        """
        return await self.get("/register", headers={"Authorization": auth})

    async def posts(self, begin: int = None, end: int = None) -> List[Post]:
        """
        查询博文
        """
        params = {}
        if begin is not None:
            params["begin"] = begin
        if end is not None:
            params["end"] = end
        resp, err = await self.get("/posts", params=params)
        if err is not None:
            return []
        return [Post.parse(v) for v in resp["posts"]]

    async def branches(self):
        """
        查询某博文不同版本
        """
        return await self.get("/branches/:platform/:mid")

    async def comments(self):
        """
        查询某博文的评论
        """
        return await self.get("/comments/:platform/:mid")

    async def log(self) -> Tuple[List[str], Optional[Exception | ApiException]]:
        """
        读取日志
        """
        return await self.get("/log")

    async def ping(self):
        """
        更新自身在线状态
        """
        await self.get("/ping")
    
    async def me(self) -> Tuple[User, Optional[Exception | ApiException]]:
        """
        获取自身信息
        """
        data, err = await self.get("/me")
        if err is not None:
            return User(0, 0), err
        return User(**data), None

    async def update(self):
        """
        主动更新主页
        """
        return await self.get("/update")

    async def modify(self, listen: List[str]) -> Tuple[List[str], Optional[Exception | ApiException]]:
        """
        更新监听列表
        """
        return await self.get("/modify", params={"listen": listen})

    async def add(self, job: Job) -> Tuple[List[Job], Optional[Exception | ApiException]]:
        """
        新增任务
        """
        resp, err = await self.post("/add", data=job.json)
        if err is not None:
            return None, err
        return [Job(**job) for job in resp], None

    async def remove(self, jobs: List[str]) -> Tuple[List[Job], Optional[Exception | ApiException]]:
        """
        移除任务
        """
        resp, err = await self.get("/remove", params={"jobs": jobs})
        if err is not None:
            return None, err
        return [Job(**job) for job in resp], None
    
    async def test(self, jobs: List[str]) -> Tuple[List[Job], Optional[Exception | ApiException]]:
        """
        测试任务
        """
        return await self.get("/test", params={"jobs": jobs})

    async def submit(self, post: Post):
        """
        提交博文
        """
        return (await self.post("/submit", data=post.json))[1]

    async def exec(self, cmd: str) -> Tuple[List[str], Optional[Exception | ApiException]]:
        """
        执行命令
        """
        return await self.get(f"/exec/{cmd}")

    async def users(self) -> List[User]:
        """
        获取所有用户
        """
        resp, err = await self.get("/users")
        if err is not None:
            logger.error(err)
            return []
        return [User(**u) for u in resp]

    async def permission(self, uid: str, perm: float):
        """
        修改权限
        """
        return (await self.get(f"/permission/{uid}/{perm}"))[1]

    async def close(self) -> str:
        """
        结束进程
        """
        return await self.must_get("/close")
    
    async def clear(self) -> str:
        """
        删除资源
        """
        return await self.must_get("/clear")
    
    async def reboot(self) -> str:
        """
        删库重启
        """
        return await self.must_get("/reboot")
    
    async def files(self) -> Tuple[Optional[str], Optional[Exception | ApiException]]:
        """
        检查资源
        """
        return await self.get("/files")
    


import datetime
import json
from dataclasses import dataclass, field
from typing import Dict, List


@dataclass
class Attachment:
    url: str
    local: str = ""
    MIME: str = ""

    @property
    def data(self) -> dict:
        """
        返回字典格式对象
        """
        return {"url": self.url}

    def __str__(self) -> str:
        """
        返回字典格式经 json 处理后对象
        """
        return json.dumps(self.data, ensure_ascii=False)


def Attachments(*urls: str) -> List[Attachment]:
    return [Attachment(u) for u in urls if u is not None and u != ""]


@dataclass
class Online:
    server: int
    users: Dict[str, int]


@dataclass
class Job:
    pattern: str
    method: str
    url: str
    data: Dict[str, str] = field(default_factory=dict)
    headers: Dict[str, str] = field(default_factory=dict)
    cookies: Dict[str, str] = field(default_factory=dict)
    id: int = 0

    @property
    def json(self) -> dict:
        """
        返回字典格式经 json 处理后对象
        """
        dic = {}
        for k, v in self.__dict__.items():
            if k in ["data", "headers", "cookies"]:
                dic[k] = json.dumps(v)
            else:
                dic[k] = str(v)
        return dic


@dataclass
class User:
    uid: str
    token: str = ""
    permission: float = 5.1
    jobs: List[Job] = field(default_factory=list)
    listening: List[str] = field(default_factory=list)


@dataclass
class Blogger:
    platform: str
    uid: str
    name: str
    create: str
    follower: str
    following: str
    description: str
    face: Attachment
    pendant: Attachment

    @property
    def data(self) -> dict:
        """
        返回字典格式对象
        """
        dic = {}
        for k, v in self.__dict__.items():
            if k in ["face", "pendant"]:
                v: Attachment
                dic[k] = {} if v is None else v.data
            else:
                dic[k] = str(v)
        return dic
    
    def __str__(self) -> str:
        return json.dumps(self.data, ensure_ascii=False)


@dataclass
class Post:
    mid: str
    time: str
    text: str
    source: str
    blogger: Blogger
    comments: List["Post"]
    attachments: List[Attachment]
    repost: "Post" = None
    submitter: User = None

    def set_repost(self, repost: "Post") -> "Post":
        self.repost = repost
        return self

    @property
    def date(self) -> str:
        """
        返回规定格式字符串时间
        """
        return datetime.datetime.fromtimestamp(float(self.time)).strftime("%H:%M:%S")

    @property
    def data(self) -> dict:
        """
        返回字典格式对象
        """
        dic = {}
        for k, v in self.__dict__.items():
            if k == "submitter":
                continue
            elif k == "blogger":
                dic[k] = self.blogger.data
            elif k == "attachments":
                dic[k] = [v.data for v in self.attachments]
            elif k == "repost":
                dic[k] = None if self.repost is None else self.repost.data
            elif k == "comments":
                dic[k] = [v.data for v in self.comments]
            else:
                if v is None:
                    continue
                dic[k] = str(v)
        return dic

    @property
    def json(self) -> dict:
        """
        返回字典格式经 json 处理后对象
        """
        dic = {}
        for k, v in self.__dict__.items():
            if k == "submitter":
                continue
            elif k == "attachments":
                dic[k] = [str(v) for v in self.attachments]
            elif k == "repost":
                dic[k] = "null" if self.repost is None else self.repost.string()
            elif k == "comments":
                dic[k] = [v.string() for v in self.comments]
            else:
                if v is None:
                    continue
                dic[k] = str(v)
        return dic

    def string(self) -> str:
        return json.dumps(self.data, ensure_ascii=False)
    
    @staticmethod
    def parse(post: dict) -> "Post":
        """
        递归解析
        """
        if post is None or len(post) == 0:
            return None
        user: dict = post.pop("blogger")
        return Post(
            blogger     = Blogger(
                face    = Attachment(**user.pop("face")),
                pendant = Attachment(**user.pop("pendant")),
                **user,
            ),
            comments    = [Post.parse(v) for v in post.pop("comments")],
            attachments = [Attachment(**v) for v in post.pop("attachments")],
            repost      = Post.parse(post.pop("repost")),
            submitter   = User(**post.pop("submitter")),
            **post,
        )



from . import adapters, settings
from .adapters import *
from .client import Client, Submitter, delta, run_forever
from .data import Attachment, Attachments, Blogger, Job, Post, User
from .network import HEADERS, Api, ApiException, Session

__all__ = [
    "adapters",
    "Api",
    "ApiException",
    "Attachment",
    "Attachments",
    "Blogger",
    "Client",
    "delta",
    "HEADERS",
    "Job",
    "Post",
    "run_forever",
    "Session",
    "settings",
    "Submitter",
    "User",
]



import asyncio
import datetime
from typing import Coroutine, List, Optional, Tuple

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from bilibili_api import Credential, comment, live
from loguru import logger

from .network import Api, ApiException


def delta(seconds: float):
    return datetime.datetime.now() + datetime.timedelta(seconds=seconds)


def run_forever(coro: Coroutine, *args, **kwargs):
    loop = asyncio.get_event_loop()
    loop.create_task(coro(*args, **kwargs))
    loop.run_forever()


class Client(Api, Credential):
    """
    客户端层
    
    对 Api 进一步封装
    """
    def __init__(
        self,
        url: str,
        auth: str | None = None,
        sessdata: str | None = None,
        bili_jct: str | None = None,
        dedeuserid: str | None = None,
    ):
        """
        `auth` 和 `dedeuserid` 可不填 若不填则自动获取 自动获取需要 `sessdata` 和 `bili_jct`

        如果 同时 填入 `auth` 和 `dedeuserid` 则不需要 `sessdata` 和 `bili_jct`

        Args:
            url: 基础接口地址
        
            auth: 鉴权码

            sessdata: 浏览器 Cookies 中的 SESSDATA 字段值. Defaults to None.

            bili_jct: 浏览器 Cookies 中的 bili_jct 字段值. Defaults to None.

            dedeuserid: 浏览器 Cookies 中的 DedeUserID 字段值. Defaults to None.
        """
        if auth is None:
            auth = ""
        Api.__init__(self, url, auth)
        Credential.__init__(self, sessdata=sessdata, bili_jct=bili_jct, dedeuserid=dedeuserid)

    async def get_self_uid(self) -> Tuple[str, Optional[Exception]]:
        """
        获取自身 uid
        """
        try:
            info = await live.get_self_info(self)
            return str(info["uid"]), None
        except Exception as e:
            return "", e

    async def get_self_auth(self) -> Tuple[str, Optional[Exception | ApiException]]:
        """
        获取自身鉴权码
        """
        data, err = await self.token(self.dedeuserid)
        if err is not None:
            return "", err
        try:
            await comment.send_comment(data["token"], data["oid"], comment.CommentResourceType.DYNAMIC, credential=self)
        except Exception as e:
            return "", e
        await asyncio.sleep(2)  # wait for bilibili update comment
        return await self.register(data["auth"])

    async def login(self) -> Optional[Exception | ApiException]:
        """
        登录
        """
        user, _ = await self.me()
        if self.dedeuserid is None:
            self.dedeuserid, err = await self.get_self_uid()
            if err is not None:
                return err
        if user.uid != self.dedeuserid:
            auth, err = await self.get_self_auth()
            if err is not None:
                return err
            self.set_auth(auth)
            user, _ = await self.me()
            if user.uid != self.dedeuserid:
                return ApiException("/login", 400, "Login failed.")
        return None


class Submitter(Client):
    """
    提交器

    use

    ```python
    async def main():
        async with Submitter(url, auth) as sub:
            @sub.job(0, 1)
            async def _():
                pass
    
    run_forever(main)
    ```

    or

    ```
    @Submitter(url, auth)
    async def _(sub: Submitter):
        @sub.job(0, 1)
        async def _():
            pass
    ```
    """
    async def __aenter__(self):
        self.__vaild = True
        self.__once: List[Coroutine] = []
        self.__scheduler = AsyncIOScheduler(timezone="Asia/Shanghai", event_loop=asyncio.get_running_loop())
        # 判断参数是否齐全
        if not all([self.auth, self.dedeuserid]) and not all([self.sessdata, self.bili_jct]):
            return await self.__aexit__(Exception, "请提供完整参数", None)
        # 模拟登录
        e = await self.login()
        if e is not None:
            return await self.__aexit__(e.__class__, str(e), None)
        return self

    async def __aexit__(self, typ_, value, trace):
        if not self.__vaild:
            return
        if typ_ is None:
            await self.__run__()
            return
        self.__vaild = False
        logger.error(f"{typ_.__name__}: {value}")
        asyncio.get_event_loop().stop()

    async def __run__(self):
        logger.info(f"auth: {self.auth}")
        logger.info(f"version: {await self.version()}")
        if len(self.__once) != 0:
            await asyncio.wait(self.__once)
        if len(self.__scheduler.get_jobs()) == 0:
            asyncio.get_event_loop().stop()
        else:
            self.__scheduler.start()

    def __call__(self, fn: Coroutine):
        async def main():
            if await self.__aenter__() is None:
                return
            try:
                await fn(self)
                await self.__run__()
            except Exception as e:
                await self.__aexit__(e.__class__, str(e), None)
        run_forever(main)

    def add_once(self, fn: Coroutine, *args, **kwargs):
        """
        添加一次任务
        """
        self.__once.append(asyncio.create_task(fn(*args, **kwargs)))
        return fn

    def once(self, *args, **kwargs):
        """
        一次任务装饰器
        """
        return lambda fn: self.add_once(fn, *args, **kwargs)

    def add_job(self, fn: Coroutine, start: int = 0, interval: int = 5, *args, **kwargs):
        """
        新增任务
        """
        self.__scheduler.add_job(fn, "interval", next_run_time=delta(start), seconds=interval, args=args, kwargs=kwargs)
        return fn

    def job(self, start: int = 0, interval: int = 5, *args, **kwargs):
        """
        新增任务装饰器
        """
        return lambda fn: self.add_job(fn, start=start, interval=interval, *args, **kwargs)



import time
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any

from loguru import logger

from ..data import Attachment, Attachments, Blogger, Post
from ..network import ApiException, Session


def created_at(text: str) -> int:
    """
    解析微博时间字段转为时间戳
    """
    return int(datetime.strptime(text, "%a %b %d %H:%M:%S %z %Y").timestamp())


def created_at_comment(text: str) -> int:
    """
    标准化微博发布时间

    参考 https://github.com/Cloud-wish/Dynamic_Monitor/blob/main/main.py#L575
    """
    created_at = datetime.now()
    
    if u"分钟" in text:
        minute = text[:text.find(u"分钟")]
        minute = timedelta(minutes=int(minute))
        created_at -= minute
    elif u"小时" in text:
        hour = text[:text.find(u"小时")]
        hour = timedelta(hours=int(hour))
        created_at -= hour
    elif u"昨天" in text:
        created_at -= timedelta(days=1)
    elif text.count('-') != 0:
        if text.count('-') == 1:
            text = f"{created_at.year}-{text}"
        created_at = datetime.strptime(text, "%Y-%m-%d")
    return int(created_at.timestamp())


def deep_get(dic: dict, keys: str, default: Any = None) -> Any:
    """
    深度获取元素
    """
    kList = keys.split(".")
    for key in kList[:-1]:
        dic = dic.get(key, {})
    return dic.get(kList[-1], default)


@dataclass
class WeiboPost(Post):
    isTop: bool = False

    @staticmethod
    def parse(mblog: dict) -> "WeiboPost":
        """
        递归解析
        """
        if mblog is None or len(mblog) == 0:
            return None
        user: dict = mblog.get("user", {})
        return WeiboPost(
            mid     = mblog["mid"],
            time    = str(created_at(mblog["created_at"])),
            text    = mblog["text"],
            source  = mblog["source"],
            blogger = Blogger(
                platform    = "weibo",
                uid         = str(user.get("id")),
                name        = user.get("screen_name"),
                create      = str(int(time.time())),
                follower    = str(user.get("followers_count")),
                following   = str(user.get("follow_count")),
                description = user.get("description"),
                face        = Attachment(user.get("avatar_hd")),
                pendant     = None,
            ),
            repost      = WeiboPost.parse(mblog.get("retweeted_status")),
            comments    = [],
            attachments = [Attachment(p["large"]["url"]) for p in mblog.get("pics", [])],
            isTop       = deep_get(mblog, "title.text", "") == "置顶",
        )


def comment(com: dict) -> Post:
    """
    解析评论
    """
    if com is None or len(com) == 0:
        return None
    user: dict = com.get("user", {})
    return Post(
        mid     = str(com["id"]),
        time    = str(created_at_comment(com["created_at"])),
        text    = com["text"],
        source  = com["source"],
        blogger = Blogger(
            platform    = "weiboComment",
            uid         = str(user.get("id")),
            name        = user.get("screen_name"),
            create      = str(int(time.time())),
            follower    = str(user.get("followers_count")),
            following   = str(user.get("friends_count")),
            description = user.get("description", ""),
            face        = Attachment(user.get("profile_image_url")),
            pendant     = None,
        ),
        comments    = [],
        attachments = Attachments(deep_get(com, "pic.large.url")),
    )


class Weibo(Session):
    """
    微博适配器
    """
    def __init__(self, url: str = "https://m.weibo.cn/api", headers: dict | None = None) -> None:
        super().__init__(url, headers)

    def check_code(self, url: str, result: Any) -> ApiException | None:
        if result["ok"] != 1:
            return ApiException(url=url, code=result["ok"], **result)
        return None

    def data(self, result: Any) -> Any:
        return result["data"]

    async def posts(self, uid: str, page: int = 1):
        data, err = await self.get(f"/container/getIndex?containerid=107603{uid}&page={page}")
        if err is not None:
            logger.error(err)
            return
        for card in data["cards"]:
            if card["card_type"] != 9:
                continue
            yield WeiboPost.parse(card["mblog"])

    async def comments(self, post: Post):
        data, err = await self.get(f"/comments/show?id={post.mid}")
        if err is not None:
            logger.error(err)
            return
        for com in data["data"]:
            yield comment(com).set_repost(post)

        # 清空储存的评论
        # postName = post.type + post.mid
        # if self.usersComments.get(post.uid, None) is None:
        #     self.usersComments[post.uid] = Comments(postName)
        # elif self.usersComments[post.uid].post != postName:
        #     self.usersComments[post.uid] = Comments(postName)

        
            # mid = str(com["id"])
            # if mid in self.usersComments[post.uid].comments: continue

        #     com["attachment"] = [postName]
        #     com["repost"] = self.usersComments[post.uid].comments.get(str(com.get("reply_id", "")), None)
        #     self.usersComments[post.uid].comments[mid] = com
        #     yield WeiboComment.parse(com)